"""Utility functions for hyfi"""
import ast
import itertools
import re
import sys
from contextlib import suppress
from typing import Any, Dict, List, Optional, TextIO, Type, Union

import chardet
import colorama
from pydantic import StrictBool

from hyfi.utils.types import IntSeq

colorama.init()


class Style:
    """Common color styles."""

    OK: IntSeq = [colorama.Fore.GREEN, colorama.Style.BRIGHT]  # type: ignore
    WARNING: IntSeq = [colorama.Fore.YELLOW, colorama.Style.BRIGHT]  # type: ignore
    IGNORE: IntSeq = [colorama.Fore.CYAN]  # type: ignore
    DANGER: IntSeq = [colorama.Fore.RED, colorama.Style.BRIGHT]  # type: ignore
    RESET: IntSeq = [colorama.Fore.RESET, colorama.Style.RESET_ALL]  # type: ignore


INDENT = " " * 2
HLINE = "-" * 42

NO_VALUE: object = object()


class FUNCs:
    @staticmethod
    def unescape_dict(d):
        """Unescape a dictionary"""
        return ast.literal_eval(repr(d).encode("utf-8").decode("unicode-escape"))

    @staticmethod
    def lower_case_with_underscores(string):
        """Converts 'CamelCased' to 'camel_cased'."""
        return (
            re.sub(r"\s+", "_", re.sub(r"(?<!^)(?=[A-Z])", "_", string).lower())
            .replace("-", "_")
            .replace("__", "_")
        )

    @staticmethod
    def ordinal(num):
        """Return the ordinal of a number as a string."""
        return "%d%s" % (
            num,
            "tsnrhtdd"[(num // 10 % 10 != 1) * (num % 10 < 4) * num % 10 :: 4],
        )

    @staticmethod
    def get_offset_ranges(count, num_workers):
        """Get offset ranges for parallel processing"""
        assert count > num_workers
        step_sz = int(count / num_workers)
        offset_ranges = [0]
        pv_cnt = 1
        for i in range(num_workers):
            pv_cnt = count + 1 if i == num_workers - 1 else pv_cnt + step_sz
            offset_ranges.append(pv_cnt)
        return offset_ranges

    @staticmethod
    def fancy_print(*args, color=None, bold=False, **kwargs):
        """Print with color and bold"""
        if bold:
            print("\033[1m", end="")

        if color:
            print(f"\033[{color}m", end="")

        print(*args, **kwargs)

        print("\033[0m", end="")  # reset

    # https://stackoverflow.com/questions/12523586/python-format-size-application-converting-b-to-kb-mb-gb-tb/37423778
    @staticmethod
    def humanbytes(B, units=None):
        "Return the given bytes as a human friendly KB, MB, GB, or TB string"
        B = float(B)
        KB = float(1024)
        MB = float(KB**2)  # 1,048,576
        GB = float(KB**3)  # 1,073,741,824
        TB = float(KB**4)  # 1,099,511,627,776

        if (B < KB and units is None) or units == "B":
            return "{0} {1}".format(B, "Bytes" if 0 == B > 1 else "Byte")
        elif (KB <= B < MB and units is None) or units == "KiB":
            return "{0:.2f} KiB".format(B / KB)
        elif (MB <= B < GB and units is None) or units == "MiB":
            return "{0:.2f} MiB".format(B / MB)
        elif (GB <= B < TB and units is None) or units == "GiB":
            return "{0:.2f} GiB".format(B / GB)
        elif (TB <= B and units is None) or units == "TiB":
            return "{0:.2f} TiB".format(B / TB)

    @staticmethod
    def parse_size(sizestr):
        """
        Parse a size string into a number of bytes. For example, "16K" will
        return 16384.  If no suffix is provided, bytes are assumed.  This
        function is case-insensitive.

        :param sizestr: A string representing a size, such as "16K", "2M", "1G".
        :return: The number of bytes that the string represents.
        """
        unit = sizestr[-1]
        size = float(sizestr[:-1])

        if unit.upper() == "B":
            return size
        if unit.upper() == "K":
            return size * 1024
        if unit.upper() == "M":
            return size * 1024 * 1024
        if unit.upper() == "G":
            return size * 1024 * 1024 * 1024
        if unit.upper() == "T":
            return size * 1024 * 1024 * 1024 * 1024

    @staticmethod
    def check_min_len(s, len_func, min_len):
        """Check if the length of a string is greater than or equal to a minimum length"""
        return len_func(s) >= min_len

    @staticmethod
    def check_max_len(s, len_func, max_len):
        """Check if the length of a string is less than or equal to a maximum length"""
        return len_func(s) <= max_len

    @staticmethod
    def utf8len(s):
        """Return the length of a string in bytes"""
        return len(str(s).encode("utf-8"))

    @staticmethod
    def len_wospc(x):
        """Return the length of a string in bytes without spaces"""
        return FUNCs.utf8len(re.sub(r"\s", "", str(x)))

    @staticmethod
    def len_bytes(x):
        """Return the length of a string in bytes"""
        return FUNCs.utf8len(x)

    @staticmethod
    def len_words(x):
        """Return the number of words in a string"""
        return len(x.split()) if isinstance(x, str) else 0

    @staticmethod
    def len_sents(x, sep):
        """Return the number of sentences in a string"""
        sep = str(sep).encode("utf-8").decode("unicode-escape")
        return len(re.sub(r"(\r?\n|\r){1,}", sep, x).split(sep))

    @staticmethod
    def len_segments(x, sep):
        """Return the number of segments in a string"""
        sep = str(sep).encode("utf-8").decode("unicode-escape")
        return len(re.sub(r"(\r?\n|\r){2,}", sep, x).split(sep))

    @staticmethod
    def any_to_utf8(b):
        """Convert any string to utf-8"""
        try:
            return b.decode("utf-8")
        except UnicodeDecodeError:
            # try to figure out encoding if not utf-8

            guess = chardet.detect(b)["encoding"]

            if not guess or guess == "UTF-8":
                return

            try:
                return b.decode(guess)
            except (UnicodeDecodeError, LookupError):
                # still cant figure out encoding, give up
                return

    @staticmethod
    def today(_format="%Y-%m-%d"):
        """Return today's date"""
        from datetime import datetime

        if _format is None:
            return datetime.today().date()
        else:
            return datetime.today().strftime(_format)

    @staticmethod
    def now(_format="%Y-%m-%d %H:%M:%S"):
        """Return current date and time"""
        from datetime import datetime

        return datetime.now() if _format is None else datetime.now().strftime(_format)

    @staticmethod
    def strptime(
        _date_str: str,
        _format: str = "%Y-%m-%d",
    ):
        """Return a datetime object from a string"""
        from datetime import datetime

        return datetime.strptime(_date_str, _format)

    @staticmethod
    def to_dateparm(_date, _format="%Y-%m-%d"):
        """Return a date parameter string"""
        from datetime import datetime

        _dtstr = datetime.strftime(_date, _format)
        _dtstr = "${to_datetime:" + _dtstr + "," + _format + "}"
        return _dtstr

    @staticmethod
    def human_readable_type_name(t: Type) -> str:
        """
        Generates a useful-for-humans label for a type.
        For builtin types, it's just the class name (eg "str" or "int").
        For other types, it includes the module (eg "pathlib.Path").
        """
        module = t.__module__
        if module == "builtins":
            return t.__qualname__
        elif module.split(".")[0] == "hyfi":
            module = "hyfi"

        try:
            return f"{module}.{t.__qualname__}"
        except AttributeError:
            return str(t)

    @staticmethod
    def readable_types_list(type_list: List[Type]) -> str:
        """Generates a useful-for-humans label for a list of types."""
        return ", ".join(FUNCs.human_readable_type_name(t) for t in type_list)

    @staticmethod
    def dict_product(dicts) -> List[Dict]:
        """
        >>> list(dict_product(dict(number=[1,2], character='ab')))
        [{'character': 'a', 'number': 1},
        {'character': 'a', 'number': 2},
        {'character': 'b', 'number': 1},
        {'character': 'b', 'number': 2}]
        """
        return [dict(zip(dicts, x)) for x in itertools.product(*dicts.values())]

    @staticmethod
    def printf(
        action: str,
        msg: Any = "",
        style: Optional[IntSeq] = None,
        indent: int = 10,
        verbose: Union[bool, StrictBool] = True,
        file_: TextIO = sys.stdout,
    ) -> Optional[str]:
        """Print string with common format."""
        if not verbose:
            return None  # HACK: Satisfy MyPy
        _msg = str(msg)
        action = action.rjust(indent, " ")
        if not style:
            return action + _msg

        out = style + [action] + Style.RESET + [INDENT, _msg]  # type: ignore
        print(*out, sep="", file=file_)
        return None  # HACK: Satisfy MyPy

    @staticmethod
    def printf_exception(
        e: Exception, action: str, msg: str = "", indent: int = 0, quiet: bool = False
    ) -> None:
        """Print exception with common format."""
        if not quiet:
            print("", file=sys.stderr)
            FUNCs.printf(
                action, msg=msg, style=Style.DANGER, indent=indent, file_=sys.stderr
            )
            print(HLINE, file=sys.stderr)
            print(e, file=sys.stderr)
            print(HLINE, file=sys.stderr)

    @staticmethod
    def cast_str_to_bool(value: Any) -> bool:
        """Parse anything to bool.

        Params:
            value:
                Anything to be casted to a bool. Tries to be as smart as possible.

                1.  Cast to number. Then: 0 = False; anything else = True.
                1.  Find [YAML booleans](https://yaml.org/type/bool.html),
                    [YAML nulls](https://yaml.org/type/null.html) or `none` in it
                    and use it appropriately.
                1.  Cast to boolean using standard python `bool(value)`.
        """
        # Assume it's a number
        with suppress(TypeError, ValueError):
            return bool(float(value))
        # Assume it's a string
        with suppress(AttributeError):
            lower = value.lower()
            if lower in {"y", "yes", "t", "true", "on"}:
                return True
            elif lower in {"n", "no", "f", "false", "off", "~", "null", "none"}:
                return False
        # Assume nothing
        return bool(value)

    @staticmethod
    def force_str_end(original_str: str, end: str = "\n") -> str:
        """Make sure a `original_str` ends with `end`.

        Params:
            original_str: String that you want to ensure ending.
            end: String that must exist at the end of `original_str`
        """
        return original_str if original_str.endswith(end) else original_str + end



"""
    HiFY Base Config
"""
from pathlib import Path
from typing import Any, Dict, List, Optional, Set, Union

from omegaconf import DictConfig
from pydantic import BaseModel, ConfigDict, model_validator

from hyfi.composer import Composer
from hyfi.utils.logging import LOGGING

logger = LOGGING.getLogger(__name__)


class BaseConfig(BaseModel):
    """
    Base class for all config classes.
    """

    _config_name_: str = "__init__"
    _config_group_: str = ""
    verbose: bool = False

    _init_args_: Dict[str, Any] = {}
    _exclude_: Set[str] = set()
    _property_set_methods_: Dict[str, str] = {}
    _subconfigs_: Dict[str, Any] = {}

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        extra="allow",
        validate_assignment=False,
    )  # type: ignore

    def __init__(self, **config_kwargs):
        logger.debug(
            "init %s with %s args", self.__class__.__name__, len(config_kwargs)
        )
        super().__init__(**config_kwargs)
        self.initialize_subconfigs(config_kwargs)

    def __setattr__(self, key, val):
        """
        Overrides the default __setattr__ method to allow for custom property set methods.

        Args:
            key (str): The name of the attribute to set.
            val (Any): The value to set the attribute to.
        """
        if method := self._property_set_methods_.get(key):  # type: ignore
            logger.info(
                "Setting %s to %s",
                key,
                val if isinstance(val, (str, int)) else type(val),
            )
            getattr(self, method)(val)
        super().__setattr__(key, val)

    @model_validator(mode="before")
    def validate_model_config_before(cls, data):
        # logger.debug("Validating model config before validating each field.")
        _config_name_ = data.get("_config_name_", getattr(cls._config_name_, "default", "__init__"))  # type: ignore
        _config_group_ = data.get("_config_group_", getattr(cls._config_group_, "default"))  # type: ignore
        _class_name_ = cls.__name__  # type: ignore
        if not _config_group_:
            logger.debug("There is no config group specified.")
            return data
        # Initialize the config with the given config_name.
        logger.info(
            "Composing `%s` class with `%s` config in `%s` group.",
            _class_name_,
            _config_name_,
            _config_group_,
        )
        config_group = f"{_config_group_}={_config_name_}"
        cfg = Composer(
            config_group=config_group,
            config_data=data,
            throw_on_compose_failure=False,
        ).config_as_dict
        data = Composer.update(cfg, data)
        # Exclude any attributes specified in the class's `exclude` list.
        exclude = getattr(cls._exclude_, "default", set())  # type: ignore
        for name in exclude:
            if name in data:
                del data[name]  # type: ignore
        return data

    # @model_validator(mode="after")  # type: ignore
    # def validate_model_config_after(cls, model):
    #     logger.debug("validate_model_config_after")
    #     return model

    def initialize_subconfigs(self, config_kwargs: Dict[str, Any]):
        """
        Initializes subconfigs with the given config data.
        The function updates the object's dictionary with the given config data,
        after excluding any attributes specified in the object's `exclude` list.

        Args:
            config_kwargs (Dict[str, Any]): The config data to initialize the subconfigs with.

        Returns:
            None
        """
        self._subconfigs_ = self._subconfigs_ or {}
        for name, config in self._subconfigs_.items():
            if name in config_kwargs and isinstance(config_kwargs[name], dict):
                cfg = config_kwargs[name]
                logger.debug("Initializing subconfig %s with %s", name, cfg)
                setattr(self, name, config.model_validate(cfg))

    def export_config(
        self,
        exclude: Optional[Union[str, List[str], Set[str], None]] = None,
        exclude_none: bool = True,
        only_include: Optional[Union[str, List[str], Set[str], None]] = None,
    ) -> Dict[str, Any]:
        """
        Export the configuration to a dictionary.

        Args:
            exclude (Optional[Union[str, List[str], Set[str], None]]): Keys to exclude from the saved configuration.
                Defaults to None.
            exclude_none (bool): Whether to exclude keys with None values. Defaults to True.
            only_include (Optional[Union[str, List[str], Set[str], None]]): Keys to include in the saved configuration.
                Defaults to None.

        Returns:
            Dict[str, Any]: The configuration dictionary.
        """
        if not exclude:
            exclude = self._exclude_  # type: ignore
        if isinstance(exclude, str):
            exclude = [exclude]
        if exclude is None:
            exclude = []
        if isinstance(only_include, str):
            only_include = [only_include]
        if only_include is None:
            only_include = []

        config = self.model_dump(exclude=exclude, exclude_none=exclude_none)  # type: ignore
        if only_include:
            config = {key: config[key] for key in only_include if key in config}

        return config

    def save_config(
        self,
        filepath: Union[str, Path],
        exclude: Optional[Union[str, List[str], Set[str], None]] = None,
        exclude_none: bool = True,
        only_include: Optional[Union[str, List[str], Set[str], None]] = None,
    ) -> str:
        """
        Save the batch configuration to file.

        Args:
            filepath ([Union[str, Path]): The filepath to save the configuration to.
            exclude (Optional[Union[str, List[str], Set[str], None]]): Keys to exclude from the saved configuration.
                Defaults to None.
            exclude_none (bool): Whether to exclude keys with None values. Defaults to True.
            only_include (Optional[Union[str, List[str], Set[str], None]]): Keys to include in the saved configuration.
                Defaults to None.

        Returns:
            str: The filename of the saved configuration.
        """
        logger.info("Saving config to %s", filepath)

        config_to_save = self.export_config(
            exclude=exclude, exclude_none=exclude_none, only_include=only_include
        )

        Composer.save(config_to_save, filepath)
        return str(filepath)

    def save_config_as_json(
        self,
        filepath: Union[str, Path],
        exclude: Optional[Union[str, List[str], Set[str], None]] = None,
        exclude_none: bool = True,
        only_include: Optional[Union[str, List[str], Set[str], None]] = None,
    ) -> str:
        def dumper(obj):
            return Composer.to_dict(obj) if isinstance(obj, DictConfig) else str(obj)

        config_to_save = self.export_config(
            exclude=exclude, exclude_none=exclude_none, only_include=only_include
        )
        logger.info("Saving config to %s", filepath)
        Composer.save_json(config_to_save, filepath, default=dumper)
        return str(filepath)

    def print_config(
        self,
    ):
        Composer.print(self.model_dump())



"""
    Hydra configuration management
"""
import collections.abc
import os
import re
from enum import Enum
from typing import Any, Callable, Dict, List, Mapping, Optional, Set, Tuple, Union

import hydra
from hydra.core.global_hydra import GlobalHydra
from omegaconf import DictConfig
from pydantic import BaseModel, ConfigDict, PrivateAttr

from hyfi.core import (
    __about__,
    __hydra_config__,
    __hydra_default_config_group_value__,
    __hydra_version_base__,
)
from hyfi.core import hydra as hyfi_hydra
from hyfi.utils.conf import CONFs
from hyfi.utils.logging import LOGGING
from hyfi.utils.packages import PKGs

if level := os.environ.get("HYFI_LOG_LEVEL"):
    LOGGING.setLogger(level)
logger = LOGGING.getLogger(__name__)

__global_package_list__: Set[str] = {"cmd", "mode", "workflow"}


class SpecialKeys(str, Enum):
    """Special keys in configs used by HyFI."""

    CALL = "_call_"
    CONFIG = "_config_"
    CONFIG_GROUP = "_config_group_"
    CONFIG_NAME = "_config_name_"
    EXEC = "_exec_"
    FUNC = "_func_"
    KWARGS = "_kwargs_"
    METHOD = "_method_"
    PARTIAL = "_partial_"
    PIPE = "_pipe_"
    RECURSIVE = "_recursive_"
    RUN = "_run_"
    TARGET = "_target_"
    TYPE = "_type_"
    WITH = "_with_"


class Composer(BaseModel, CONFs):
    """
    Compose a configuration by applying overrides
    """

    config_group: Optional[str] = None
    overrides: Optional[List[str]] = None
    config_data: Optional[Union[Dict[str, Any], DictConfig]] = None
    throw_on_resolution_failure: bool = True
    throw_on_missing: bool = False
    config_name: Optional[str] = None
    config_module: Optional[str] = None
    global_package: bool = False
    verbose: bool = False

    _cfg_: DictConfig = PrivateAttr({})

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        validate_assignment=True,
    )  # type: ignore

    def __init__(self, **args):
        super().__init__(**args)
        self._cfg_ = self.compose(
            config_group=self.config_group,
            overrides=self.overrides,
            config_data=self.config_data,
            throw_on_resolution_failure=self.throw_on_resolution_failure,
            throw_on_missing=self.throw_on_missing,
            root_config_name=self.config_name,
            config_module=self.config_module,
            global_package=self.global_package,
            verbose=self.verbose,
        )

    def compose(
        self,
        config_group: Optional[str] = None,
        overrides: Optional[List[str]] = None,
        config_data: Optional[Union[Dict[str, Any], DictConfig]] = None,
        throw_on_compose_failure: bool = True,
        throw_on_resolution_failure: bool = True,
        throw_on_missing: bool = False,
        root_config_name: Optional[str] = None,
        config_module: Optional[str] = None,
        global_package: bool = False,
        verbose: bool = False,
    ) -> DictConfig:
        """
        Compose a configuration by applying overrides

        Args:
            config_group: Name of the config group to compose (`config_group=name`)
            overrides: List of config groups to apply overrides to (`overrides=["override_name"]`)
            config_data: Keyword arguments to override config group values (will be converted to overrides of the form `config_group.key=value`)
            return_as_dict: Return the result as a dict
            throw_on_compose_failure: If True throw an exception if composition fails
            throw_on_resolution_failure: If True throw an exception if resolution fails
            throw_on_missing: If True throw an exception if config_group doesn't exist
            root_config_name: Name of the root config to be used (e.g. `hconf`)
            config_module: Module of the config to be used (e.g. `hyfi.conf`)
            global_package: If True, the config assumed to be a global package
            verbose: If True print configuration to stdout

        Returns:
            A config object or a dictionary with the composed config
        """
        self._cfg_ = Composer._compose(
            config_group=config_group,
            overrides=overrides,
            config_data=config_data,
            throw_on_compose_failure=throw_on_compose_failure,
            throw_on_resolution_failure=throw_on_resolution_failure,
            throw_on_missing=throw_on_missing,
            root_config_name=root_config_name,
            config_module=config_module,
            global_package=global_package,
            verbose=verbose,
        )
        return self._cfg_

    @property
    def config(self) -> DictConfig:
        """
        Returns the composed configuration.
        """
        return self._cfg_

    @property
    def config_as_dict(self) -> Dict:
        """
        Return the configuration as a dictionary.
        """
        return Composer.to_dict(self._cfg_)

    def __call__(
        self,
        config_group: Optional[str] = None,
        overrides: Optional[List[str]] = None,
        config_data: Optional[Union[Dict[str, Any], DictConfig]] = None,
        throw_on_compose_failure: bool = True,
        throw_on_resolution_failure: bool = True,
        throw_on_missing: bool = False,
        root_config_name: Optional[str] = None,
        config_module: Optional[str] = None,
        global_package: bool = False,
        verbose: bool = False,
    ) -> DictConfig:
        return self.compose(
            config_group=config_group,
            overrides=overrides,
            config_data=config_data,
            throw_on_compose_failure=throw_on_compose_failure,
            throw_on_resolution_failure=throw_on_resolution_failure,
            throw_on_missing=throw_on_missing,
            root_config_name=root_config_name,
            config_module=config_module,
            global_package=global_package,
            verbose=verbose,
        )

    def __getitem__(self, key):
        return self._cfg_[key]

    def __iter__(self):
        return iter(self._cfg_)

    def __len__(self):
        return len(self._cfg_)

    def __contains__(self, key):
        return key in self._cfg_

    def __eq__(self, other):
        return self._cfg_ == other

    def __ne__(self, other):
        return self._cfg_ != other

    def __bool__(self):
        return bool(self._cfg_)

    def __hash__(self):
        return hash(self._cfg_)

    def __getstate__(self):
        return self._cfg_

    @staticmethod
    def hydra_compose(
        root_config_name: Optional[str] = None,
        config_module: Optional[str] = None,
        overrides: Optional[List[str]] = None,
    ):
        is_initialized = GlobalHydra.instance().is_initialized()  # type: ignore
        config_module = config_module or __hydra_config__.hyfi_config_module
        logger.debug("config_module: %s", config_module)
        if is_initialized:
            # Hydra is already initialized.
            logger.debug("Hydra is already initialized")
            cfg = hydra.compose(config_name=root_config_name, overrides=overrides)
        else:
            with hyfi_hydra.initialize_config(
                config_module=config_module,
                config_dir=__hydra_config__.hyfi_user_config_path,
                version_base=__hydra_version_base__,
            ):
                cfg = hydra.compose(config_name=root_config_name, overrides=overrides)
        return cfg

    @staticmethod
    def split_config_group(
        config_group: Optional[str] = None,
    ) -> Tuple[str, str, str]:
        if config_group:
            group_ = config_group.split("=")
            # group_key group_value group_key group_value group_key group_value default
            if len(group_) == 2:
                group_key, group_value = group_
            else:
                group_key = group_[0]
                group_value = __hydra_default_config_group_value__
            config_group = f"{group_key}={group_value}"
        else:
            group_key = ""
            group_value = ""
            config_group = ""
        return config_group, group_key, group_value

    @staticmethod
    def _compose_as_dict(
        config_group: Optional[str] = None,
        overrides: Optional[List[str]] = None,
        config_data: Optional[Union[Dict[str, Any], DictConfig]] = None,
        throw_on_compose_failure: bool = True,
        throw_on_resolution_failure: bool = True,
        throw_on_missing: bool = False,
        root_config_name: Optional[str] = None,
        config_module: Optional[str] = None,
        global_package: bool = False,
        **kwargs,
    ) -> Dict:
        return Composer.to_dict(
            Composer._compose(
                config_group=config_group,
                overrides=overrides,
                config_data=config_data,
                throw_on_compose_failure=throw_on_compose_failure,
                throw_on_resolution_failure=throw_on_resolution_failure,
                throw_on_missing=throw_on_missing,
                root_config_name=root_config_name,
                config_module=config_module,
                global_package=global_package,
                **kwargs,
            )
        )

    @staticmethod
    def _compose(
        config_group: Optional[str] = None,
        overrides: Optional[List[str]] = None,
        config_data: Optional[Union[Dict[str, Any], DictConfig]] = None,
        throw_on_compose_failure: bool = True,
        throw_on_resolution_failure: bool = True,
        throw_on_missing: bool = False,
        root_config_name: Optional[str] = None,
        config_module: Optional[str] = None,
        global_package: bool = False,
        **kwargs,
    ) -> DictConfig:
        try:
            return Composer._compose_internal(
                config_group=config_group,
                overrides=overrides,
                config_data=config_data,
                throw_on_resolution_failure=throw_on_resolution_failure,
                throw_on_missing=throw_on_missing,
                root_config_name=root_config_name,
                config_module=config_module,
                global_package=global_package,
                **kwargs,
            )
        except Exception as e:
            logger.error("Error composing config: %s", e)
            if throw_on_compose_failure:
                raise e
            return DictConfig(config_data) if config_data else DictConfig({})

    @staticmethod
    def _compose_internal(
        config_group: Optional[str] = None,
        overrides: Optional[List[str]] = None,
        config_data: Optional[Union[Dict[str, Any], DictConfig]] = None,
        throw_on_resolution_failure: bool = True,
        throw_on_missing: bool = False,
        root_config_name: Optional[str] = None,
        config_module: Optional[str] = None,
        global_package: bool = False,
        **kwargs,
    ) -> DictConfig:
        if isinstance(config_data, DictConfig):
            logger.debug("returning config_data without composing")
            return config_data
        # Set overrides to the empty list if None
        if overrides is None:
            overrides = []
        # Set the group key and value of the config group.
        config_group, group_key, group_value = Composer.split_config_group(config_group)
        if group_key in __global_package_list__:
            global_package = True
        # If group_key and group_value are specified in the configuration file.
        if group_key and group_value:
            # Initialize hydra configuration module.
            cfg = Composer.hydra_compose(
                root_config_name=root_config_name,
                config_module=config_module,
                overrides=overrides,
            )
            cfg = Composer.select(
                cfg,
                key=group_key,
                default=None,
                throw_on_missing=False,
                throw_on_resolution_failure=False,
            )
            override = config_group if cfg is not None else f"+{config_group}"
            # Add override to overrides list.
            if isinstance(override, str):
                if overrides:
                    overrides.append(override)
                else:
                    overrides = [override]

        logger.debug(f"compose config with overrides: {overrides}")
        # Initialize hydra and return the configuration.
        cfg = Composer.hydra_compose(
            root_config_name=root_config_name,
            config_module=config_module,
            overrides=overrides,
        )
        # Add config group overrides to overrides list.
        if group_key and not global_package:
            group_overrides: List[str] = []
            group_cfg = Composer.select(
                cfg,
                key=group_key,
                default=None,
                throw_on_missing=False,
                throw_on_resolution_failure=False,
            )
            if config_data and group_cfg:
                group_overrides.extend(
                    f"{group_key}.{k}={v}"
                    for k, v in config_data.items()
                    if isinstance(v, (str, int, float, bool)) and k in group_cfg
                )
            if group_overrides:
                overrides.extend(group_overrides)
                cfg = Composer.hydra_compose(
                    root_config_name=root_config_name,
                    config_module=config_module,
                    overrides=overrides,
                )

            # Select the group_key from the configuration.
            cfg = Composer.select(
                cfg,
                key=group_key,
                default=None,
                throw_on_missing=throw_on_missing,
                throw_on_resolution_failure=throw_on_resolution_failure,
            )
        return cfg

    @staticmethod
    def is_composable(
        config_group: str,
        config_module: Optional[str] = None,
    ) -> bool:
        """
        Determines whether the input configuration object is composable.

        Args:
            config_group (str): The name of the configuration group to check.
            config_module (Optional[str], optional): The name of the configuration module to check. Defaults to None.

        Returns:
            bool: True if the configuration object is composable, False otherwise.
        """
        try:
            cfg = Composer._compose(
                config_group=config_group,
                config_module=config_module,
            )
            return cfg is not None
        except Exception as e:
            logger.error("Error composing config: %s", e)
            return False

    @staticmethod
    def is_instantiatable(cfg: Any):
        """
        Determines whether the input configuration object is instantiatable.

        Args:
            cfg (Any): The configuration object to check.

        Returns:
            bool: True if the configuration object is instantiatable, False otherwise.
        """
        return Composer.is_config(cfg) and SpecialKeys.TARGET in cfg

    @staticmethod
    def replace_special_keys(_dict: Mapping[str, Any]) -> Mapping:
        """
        Replace special keys in a dictionary.

        Args:
            _dict (Mapping[str, Any]): The dictionary to update.

        Returns:
            Mapping: The updated dictionary.
        """
        _new_dict = {}
        for k, v in _dict.items():
            key = Composer.generate_alias_for_special_keys(k)
            if isinstance(v, collections.abc.Mapping):
                _new_dict[key] = Composer.replace_special_keys(v)
            else:
                _new_dict[key] = v
        return _new_dict

    @staticmethod
    def generate_alias_for_special_keys(key: str) -> str:
        """
        Generate an alias for special keys.
        _with_ -> run_with
        _pipe_ -> run_pipe
        _run_ -> run

        Args:
            key (str): The special key to generate an alias for.

        Returns:
            str: The alias for the special key.
        """
        # replace the exact `with`, `pipe` with `run_with`, `run_pipe`
        key_ = re.sub(r"^with$", "run_with", key)
        # replace the prefix `_` with `run_`
        key_ = re.sub(r"^_with_$", "run_with", key_)
        key_ = re.sub(r"^_pipe_$", "pipe_target", key_)
        key_ = re.sub(r"^_run_$", "run", key_)
        return key_

    @staticmethod
    def partial(
        config: Union[str, Dict],
        *args: Any,
        **kwargs: Any,
    ) -> Callable:
        """
        Returns a callable object that is a partial version of the function or class specified in the config object.

        Args:
            config: An config object describing what to call and what params to use.
                    In addition to the parameters, the config must contain:
                    _target_ : target class or callable name (str)
                    And may contain:
                    _partial_: If True, return functools.partial wrapped method or object
                                False by default. Configure per target.
            args: Optional positional parameters pass-through
            kwargs: Optional named parameters to override
                    parameters in the config object. Parameters not present
                    in the config objects are being passed as is to the target.
                    IMPORTANT: dataclasses instances in kwargs are interpreted as config
                                and cannot be used as passthrough
        Returns:
            A callable object that is a partial version of the function or class specified in the config object.
        """
        if isinstance(config, str):
            config = {SpecialKeys.TARGET.value: config}
        else:
            config = Composer.to_dict(config)
        if not isinstance(config, dict):
            raise ValueError("config must be a dict or a str")
        config[SpecialKeys.PARTIAL.value] = True
        rc_kwargs_ = config.pop(SpecialKeys.KWARGS, {})
        if rc_kwargs_ and kwargs:
            kwargs.update(rc_kwargs_)
        return Composer.instantiate(config, *args, **kwargs)

    @staticmethod
    def instantiate(config: Any, *args: Any, **kwargs: Any) -> Any:
        """
        Instantiates an object using the provided config object.

        Args:
            config: An config object describing what to call and what params to use.
                    In addition to the parameters, the config must contain:
                    _target_ : target class or callable name (str)
                    And may contain:
                    _args_: List-like of positional arguments to pass to the target
                    _recursive_: Construct nested objects as well (bool).
                                    False by default.
                                    may be overridden via a _recursive_ key in
                                    the kwargs
                    _convert_: Conversion strategy
                            none    : Passed objects are DictConfig and ListConfig, default
                            partial : Passed objects are converted to dict and list, with
                                    the exception of Structured Configs (and their fields).
                            all     : Passed objects are dicts, lists and primitives without
                                    a trace of OmegaConf containers
                    _partial_: If True, return functools.partial wrapped method or object
                                False by default. Configure per target.
                    _args_: List-like of positional arguments
            args: Optional positional parameters pass-through
            kwargs: Optional named parameters to override
                    parameters in the config object. Parameters not present
                    in the config objects are being passed as is to the target.
                    IMPORTANT: dataclasses instances in kwargs are interpreted as config
                                and cannot be used as passthrough

        Returns:
            if _target_ is a class name: the instantiated object
            if _target_ is a callable: the return value of the call
        """
        verbose = config.get("verbose", False)
        if not Composer.is_instantiatable(config):
            if verbose:
                logger.info("Config is not instantiatable, returning config")
            return config
        _recursive_ = config.get(SpecialKeys.RECURSIVE, False)
        if SpecialKeys.RECURSIVE not in kwargs:
            kwargs[SpecialKeys.RECURSIVE.value] = _recursive_
        if verbose:
            logger.info("instantiating %s ...", config.get(SpecialKeys.TARGET))
        return hydra.utils.instantiate(config, *args, **kwargs)

    @staticmethod
    def getsource(obj: Any) -> str:
        """
        Return the source code of the object.

        Args:
            obj: The object to get the source code of.

        Returns:
            The source code of the object as a string.

        """
        try:
            target_string = ""
            if Composer.is_config(obj):
                if SpecialKeys.TARGET in obj:
                    target_string = obj[SpecialKeys.TARGET]
            elif isinstance(obj, str):
                target_string = obj
            return PKGs.getsource(target_string) if target_string else ""
        except Exception as e:
            logger.error(f"Error getting source: {e}")
            return ""

    @staticmethod
    def viewsource(obj: Any):
        """
        Print the source code of the object.

        Args:
            obj: The object to print the source code of.

        """
        print(Composer.getsource(obj))

    @staticmethod
    def instantiate_config(
        config_group: Optional[str] = None,
        overrides: Optional[List[str]] = None,
        config_data: Optional[Union[Dict[str, Any], DictConfig]] = None,
        global_package: bool = False,
        *args: Any,
        **kwargs: Any,
    ) -> Any:
        """
        Instantiates an object using the provided config group and overrides.

        Args:
            config_group: Name of the config group to compose (`config_group=name`)
            overrides: List of config groups to apply overrides to (`overrides=["override_name"]`)
            config_data: Keyword arguments to override config group values (will be converted to overrides of the form `config_group_name.key=value`)
            global_package: If True, the config assumed to be a global package
            args: Optional positional parameters pass-through
            kwargs: Optional named parameters to override
                    parameters in the config object. Parameters not present
                    in the config objects are being passed as is to the target.
                    IMPORTANT: dataclasses instances in kwargs are interpreted as config
                                and cannot be used as passthrough

        Returns:
            if _target_ is a class name: the instantiated object
            if _target_ is a callable: the return value of the call
        """
        cfg = Composer._compose(
            config_group=config_group,
            overrides=overrides,
            config_data=config_data,
            global_package=global_package,
        )
        return Composer.instantiate(cfg, *args, **kwargs)

    @staticmethod
    def print_config(
        config_group: Optional[str] = None,
        overrides: Optional[List[str]] = None,
        config_data: Optional[Union[Dict[str, Any], DictConfig]] = None,
        global_package: bool = False,
    ):
        """
        Print the configuration

        Args:
            config_group: Name of the config group to compose (`config_group=name`)
            overrides: List of config groups to apply overrides to (`overrides=["override_name"]`)
            config_data: Keyword arguments to override config group values (will be converted to overrides of the form `config_group_name.key=value`)
            global_package: If True, the config assumed to be a global package
        """
        cfg = Composer._compose(
            config_group=config_group,
            overrides=overrides,
            config_data=config_data,
            global_package=global_package,
        )
        Composer.print(cfg)



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
import functools
from textwrap import dedent
from typing import Any, Callable, List, Optional

from hydra import version
from hydra._internal.deprecation_warning import deprecation_warning
from hydra._internal.utils import get_args_parser
from hydra.core.utils import _flush_loggers
from hydra.main import _get_rerun_conf
from hydra.types import TaskFunction
from omegaconf import DictConfig

from hyfi.core.hydra.utils import _run_hydra

_UNSPECIFIED_: Any = object()


def main(
    config_path: Optional[str] = _UNSPECIFIED_,
    config_name: Optional[str] = None,
    version_base: Optional[str] = _UNSPECIFIED_,
    overrides: Optional[List[str]] = None,
) -> Callable[[TaskFunction], Any]:
    """
    :param config_path: The config path, a directory where Hydra will search for
                        config files. This path is added to Hydra's searchpath.
                        Relative paths are interpreted relative to the declaring python
                        file. Alternatively, you can use the prefix `pkg://` to specify
                        a python package to add to the searchpath.
                        If config_path is None no directory is added to the Config search path.
    :param config_name: The name of the config (usually the file name without the .yaml extension)
    """

    version.setbase(version_base)

    if config_path is _UNSPECIFIED_:
        if version.base_at_least("1.2"):
            config_path = None
        elif version_base is _UNSPECIFIED_:
            url = "https://hydra.cc/docs/1.2/upgrades/1.0_to_1.1/changes_to_hydra_main_config_path"
            deprecation_warning(
                message=dedent(
                    f"""
                config_path is not specified in @hydra.main().
                See {url} for more information."""
                ),
                stacklevel=2,
            )
            config_path = "."
        else:
            config_path = "."

    def main_decorator(task_function: TaskFunction) -> Callable[[], None]:
        @functools.wraps(task_function)
        def decorated_main(cfg_passthrough: Optional[DictConfig] = None) -> Any:
            if cfg_passthrough is not None:
                return task_function(cfg_passthrough)
            args_parser = get_args_parser()
            args = args_parser.parse_args()
            if overrides and not args.overrides:
                args.overrides = overrides
            elif overrides and args.overrides:
                args.overrides.extend(overrides)
            if args.experimental_rerun is not None:
                cfg = _get_rerun_conf(args.experimental_rerun, args.overrides)
                task_function(cfg)
                _flush_loggers()
            else:
                # no return value from run_hydra() as it may sometime actually run the task_function
                # multiple times (--multirun)
                _run_hydra(
                    args=args,
                    args_parser=args_parser,
                    task_function=task_function,
                    config_path=config_path,
                    config_name=config_name,
                )

        return decorated_main

    return main_decorator



from pathlib import Path

from pydantic import BaseModel

from hyfi.about import AboutConfig, __package_name__
from hyfi.utils.logging import LOGGING

logger = LOGGING.getLogger(__name__)

__hydra_version_base__ = "1.2"
__hydra_default_config_group_value__ = "__init__"
__config_path__ = "conf"
__config_name__ = "config"
__config_module_path__ = f"{__package_name__}.{__config_path__}"

__about__ = AboutConfig()
_batcher_instance_ = None


def __app_version__() -> str:
    """
    Returns the version of App.

    Returns:
        string containing the version of App
    """

    return __about__.__version__


def __app_path__() -> str:
    """
    Returns the path to the App root folder

    Returns:
        string containing the path to the App root folder
    """

    return __about__.__package_path__


class HydraConfig(BaseModel):
    """Global configuration for Hydra"""

    hyfi_config_path: str = __about__.config_path
    hyfi_config_module: str = __about__.config_module
    hyfi_user_config_path: str = __about__.user_config_path


__hydra_config__ = HydraConfig()


def __hyfi_path__():
    """Returns the path to the HyFI root folder"""
    return Path(__file__).parent.parent.as_posix()


def __home_path__():
    """Returns the path to the user's home folder"""
    return Path.home().as_posix()



import os
import warnings

from filelock import AcquireReturnProxy
from filelock import FileLock as _FileLock

from .common import PathOrStr


class FileLock(_FileLock):
    """
    This is just a subclass of the `FileLock` class from the `filelock` library, except that
    it adds an additional argument to the `__init__` method: `read_only_ok`.

    By default this flag is `False`, which an exception will be thrown when a lock
    can't be acquired due to lack of write permissions.
    But if this flag is set to `True`, a warning will be emitted instead of an error when
    the lock already exists but the lock can't be acquired because write access is blocked.
    """

    def __init__(
        self, lock_file: PathOrStr, timeout=-1, read_only_ok: bool = False
    ) -> None:
        super().__init__(str(lock_file), timeout=timeout)
        self._read_only_ok = read_only_ok

    def acquire(self, timeout=None, poll_interval=0.05, **kwargs) -> AcquireReturnProxy:
        try:
            return super().acquire(
                timeout=timeout, poll_interval=poll_interval, **kwargs
            )
        except OSError as err:
            # OSError could be a lot of different things, but what we're looking
            # for in particular are permission errors, such as:
            #  - errno 1  - EPERM  - "Operation not permitted"
            #  - errno 13 - EACCES - "Permission denied"
            #  - errno 30 - EROFS  - "Read-only file system"
            if err.errno not in (1, 13, 30):
                raise

            if os.path.isfile(self.lock_file) and self._read_only_ok:
                warnings.warn(
                    f"Lacking permissions required to obtain lock '{self.lock_file}'. "
                    "Race conditions are possible if other processes are writing to the same resource.",
                    UserWarning,
                )
            else:
                raise



import os
from os import PathLike
from pathlib import Path
from typing import Tuple, Union
from urllib.parse import urlparse

PathOrStr = Union[str, PathLike]

CACHE_DIRECTORY: PathOrStr = Path(
    os.getenv("CACHED_PATH_CACHE_ROOT", Path.home() / ".cache" / "cached_path")
)
"""
The default global cache directory.
"""


def _split_cloud_path(url: str, provider: str) -> Tuple[str, str]:
    """Split a full s3 path into the bucket name and path."""
    parsed = urlparse(url)
    if not parsed.netloc or not parsed.path:
        raise ValueError(f"bad {provider} path {url}")
    bucket_name = parsed.netloc
    provider_path = parsed.path
    # Remove '/' at beginning of path.
    if provider_path.startswith("/"):
        provider_path = provider_path[1:]
    return bucket_name, provider_path


def set_cache_dir(cache_dir: PathOrStr) -> None:
    """
    Set the global default cache directory.
    """
    global CACHE_DIRECTORY
    CACHE_DIRECTORY = Path(cache_dir)


def get_cache_dir() -> Path:
    """
    Get the global default cache directory.
    """
    return Path(CACHE_DIRECTORY)
